@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/child.c,v 1.2 2005/06/13 13:44:09 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/child.c,v 1.5 2005/11/09 14:12:21 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -48,8 +48,6 @@
 
 #ifdef NONE_BLOCK
 static void set_nonblock(int fd);
-#endif
-#ifdef NOT_USED
 static void unset_nonblock(int fd);
 #endif
 
@@ -59,6 +57,7 @@ static int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp);
 static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION *frontend);
 static void cancel_request(CancelPacket *sp, int secondary_backend);
 static RETSIGTYPE die(int sig);
+static RETSIGTYPE close_idle_connection(int sig);
 static int send_params(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 static void send_frontend_exits(void);
 
@@ -81,6 +80,8 @@ void do_child(int unix_fd, int inet_fd)
 	int child_idle_sec;
 	struct timeval timeout;
 	static int connected;
+	int connections_count = 0;	/* used if child_max_connections > 0 */
+
 
 	pool_debug("I am %d", getpid());
 
@@ -88,6 +89,7 @@ void do_child(int unix_fd, int inet_fd)
 	signal(SIGALRM, SIG_DFL);
 	signal(SIGTERM, die);
 	signal(SIGINT, die);
+	signal(SIGHUP, close_idle_connection);
 	signal(SIGQUIT, die);
 	signal(SIGCHLD, SIG_DFL);
 	signal(SIGUSR1, SIG_DFL);
@@ -136,7 +138,7 @@ void do_child(int unix_fd, int inet_fd)
 		/* perform accept() */
 		frontend = do_accept(unix_fd, inet_fd, &timeout);
 
-		if (frontend == NULL)
+		if (frontend == NULL)	/* connection request from frontend timed out */
 		{
 			/* check select() timeout */
 			if (connected && pool_config.child_life_time > 0 &&
@@ -149,9 +151,11 @@ void do_child(int unix_fd, int inet_fd)
 			continue;
 		}
 
-#ifdef NOT_USED
 		/* set frontend fd to blocking */
 		unset_nonblock(frontend->fd);
+
+#ifdef NOT_USED
+		set_nonblock(frontend->fd);
 #endif
 
 		/* set busy flag and clear child idle timer */
@@ -352,6 +356,18 @@ void do_child(int unix_fd, int inet_fd)
 		timeout.tv_sec = pool_config.child_life_time;
 		timeout.tv_usec = 0;
 
+		/* increment queries counter if necessary */
+		if ( pool_config.child_max_connections > 0 )
+			connections_count++;
+
+		/* check if maximum connections count for this child reached */
+		if ( ( pool_config.child_max_connections > 0 ) &&
+			( connections_count >= pool_config.child_max_connections ) )
+		{
+			pool_log("child exiting, %d connections reached", pool_config.child_max_connections);
+			send_frontend_exits();
+			exit(2);
+		}
 	}
 	exit(0);
 }
@@ -384,7 +400,6 @@ static void set_nonblock(int fd)
 }
 #endif
 
-#ifdef NOT_USED
 /*
  * unset non-block flag
  */
@@ -405,7 +420,6 @@ static void unset_nonblock(int fd)
 		exit(1);
 	}
 }
-#endif
 
 /*
 * perform accept() and return new fd
@@ -845,6 +859,7 @@ static RETSIGTYPE die(int sig)
 				pool_debug("child receives smart shutdown request but it's not in idle state");
 				return;
 			}
+			break;
 
 		case SIGINT:	/* fast shutdown */
 		case SIGQUIT:	/* immediate shutdown */
@@ -859,6 +874,37 @@ static RETSIGTYPE die(int sig)
 	exit(0);
 }
 
+static RETSIGTYPE close_idle_connection(int sig)
+{
+	int i;
+	POOL_CONNECTION_POOL *p = pool_connection_pool;
+
+	pool_debug("child receives close connection request");
+
+	for (i=0;i<pool_config.max_pool;i++, p++)
+	{
+		if (!MASTER_CONNECTION(p))
+			continue;
+		if (MASTER_CONNECTION(p)->sp->user == NULL)
+			continue;
+
+		if (MASTER_CONNECTION(p)->closetime > 0)		/* idle connection? */
+		{
+			pool_debug("close_idle_connection: close idle connection: user %s database %s", MASTER_CONNECTION(p)->sp->user, MASTER_CONNECTION(p)->sp->database);
+			pool_send_frontend_exits(p);
+			pool_free_startup_packet(MASTER_CONNECTION(p)->sp);
+			pool_close(MASTER_CONNECTION(p)->con);
+
+			if (DUAL_MODE)
+			{
+				/* do not free memory! we did not allocate them */
+				pool_close(SECONDARY_CONNECTION(p)->con);
+			}
+			memset(p, 0, sizeof(POOL_CONNECTION_POOL));
+		}
+	}
+}
+
 static void send_frontend_exits(void)
 {
 	int i;
@@ -956,6 +1002,7 @@ int health_check(void)
 		pool_error("health check failed during write. master %s at port %d is down",
 				   pool_config.current_backend_host_name,
 				   pool_config.current_backend_port);
+		close(fd);
 		return -1;
 	}
 
@@ -966,6 +1013,7 @@ int health_check(void)
 		pool_error("health check failed during write. master %s at port %d is down",
 				   pool_config.current_backend_host_name,
 				   pool_config.current_backend_port);
+		close(fd);
 		return -1;
 	}
 
@@ -989,20 +1037,22 @@ int health_check(void)
 
 	if (write(fd, &mysp, sizeof(mysp)) < 0)
 	{
-		pool_error("health check failed during write. master %s at port %d is down",
-				   pool_config.current_backend_host_name,
-				   pool_config.current_backend_port);
-		return -1;
+		pool_error("health check failed during write. secondary %s at port %d is down",
+				   pool_config.secondary_backend_host_name,
+				   pool_config.secondary_backend_port);
+		close(fd);
+		return -2;
 	}
 
 	read(fd, &kind, 1);
 
 	if (write(fd, "X", 1) < 0)
 	{
-		pool_error("health check failed during write. master %s at port %d is down",
-				   pool_config.current_backend_host_name,
-				   pool_config.current_backend_port);
-		return -1;
+		pool_error("health check failed during write. secondary %s at port %d is down",
+				   pool_config.secondary_backend_host_name,
+				   pool_config.secondary_backend_port);
+		close(fd);
+		return -2;
 	}
 
 	close(fd);
