@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/main.c,v 1.3 2005/02/25 06:38:16 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/main.c,v 1.9 2005/11/09 14:12:21 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -170,7 +170,7 @@ int main(int argc, char **argv)
 	if (pool_get_config(conf_file))
 	{
 		pool_error("Unable to get configuration. Exiting...");
-		mypid = getpid();
+		mypid = 0;		/* do not delete UNIX domain socket or pid file */
 		myexit(1);
 	}
 
@@ -183,6 +183,9 @@ int main(int argc, char **argv)
 						   (pool_config.weight_master + pool_config.weight_secondary));
 	pool_debug("weight: %ld", weight_master);
 
+	/*
+	 * if a non-switch argument remains, then it should be either "stop" or "switch"
+	 */
 	if (optind == (argc - 1))
 	{
 		if (!strcmp(argv[optind], "stop"))
@@ -195,7 +198,15 @@ int main(int argc, char **argv)
 			switch_me();
 			exit(0);
 		}
+		else
+		{
+			usage();
+			exit(1);
+		}
 	}
+	/*
+	 * else if no non-switch argument remains, then it should be a start request
+	 */
 	else if (optind == argc)
 	{
 		pid = read_pid_file();
@@ -210,13 +221,15 @@ int main(int argc, char **argv)
 				fprintf(stderr, "pid file found but it seems bogus. Trying to start pgpool anyway...\n");
 		}
 	}
-	else if (optind < argc)
+	/*
+	 * otherwise an error...
+	 */
+	else
 	{
 		usage();
 		exit(1);
 	}
 
-
 	/* set signal masks */
 	poolinitmask();
 
@@ -244,13 +257,13 @@ int main(int argc, char **argv)
 		inet_fd = create_inet_domain_socket(pool_config.listen_addresses);
 	}
 
-	pids = malloc(pool_config.num_init_children * sizeof(pool_config.num_init_children));
+	pids = malloc(pool_config.num_init_children * sizeof(pid_t));
 	if (pids == NULL)
 	{
 		pool_error("failed to allocate pids");
 		myexit(1);
 	}
-	memset(pids, 0, pool_config.num_init_children * sizeof(pool_config.num_init_children));
+	memset(pids, 0, pool_config.num_init_children * sizeof(pid_t));
 
 	/* fork the children */
 	for (i=0;i<pool_config.num_init_children;i++)
@@ -265,6 +278,7 @@ int main(int argc, char **argv)
 	pool_signal(SIGCHLD, reap_handler);
 	pool_signal(SIGUSR1, failover_handler);
 	pool_signal(SIGUSR2, failover_handler);
+	pool_signal(SIGHUP, exit_handler);
 
 	pool_log("pgpool successfully started");
 
@@ -395,7 +409,7 @@ static void stop_me(void)
 	pid = read_pid_file();
 	if (pid < 0)
 	{
-		pool_error("could read pid file");
+		pool_error("could not read pid file");
 		exit(1);
 	}
 
@@ -674,6 +688,8 @@ static RETSIGTYPE exit_handler(int sig)
 		pool_log("received fast shutdown request");
 	else if (sig == SIGQUIT)
 		pool_log("received immediate shutdown request");
+	else if (sig == SIGHUP)
+		pool_log("received idle connection close request");
 	else
 	{
 		pool_error("exit_handler: unknown signal received %d", sig);
@@ -691,14 +707,22 @@ static RETSIGTYPE exit_handler(int sig)
 			kill(pid, sig);
 		}
 	}
+
+	if (sig == SIGHUP)
+	{
+		exiting = 0;
+		POOL_SETMASK(&UnBlockSig);
+		return;
+	}
+
+	POOL_SETMASK(&UnBlockSig);
+
 	while (wait(NULL) > 0)
 		;
 
 	if (errno != ECHILD)
 		pool_error("wait() failed. reason:%s", strerror(errno));
 
-	POOL_SETMASK(&UnBlockSig);
-
 	myexit(0);
 }
 
@@ -757,7 +781,7 @@ static RETSIGTYPE failover_handler(int sig)
 	 * if not in replication mode/master slave mode, we treat this a restart request.
 	 * otherwise we need to check if we have already failovered.
 	 */
-	if (!pool_config.replication_enabled || !pool_config.master_slave_enabled ||
+	if (pool_config.replication_enabled || pool_config.master_slave_enabled ||
 		strcmp(pool_config.current_backend_host_name, pool_config.secondary_backend_host_name) ||
 		pool_config.current_backend_port != pool_config.secondary_backend_port)
 	{
@@ -773,14 +797,37 @@ static RETSIGTYPE failover_handler(int sig)
 				pool_log("starting degeneration. shutdown secondary host %s(%d)",
 						 pool_config.secondary_backend_host_name,
 						 pool_config.secondary_backend_port);
+				pool_config.server_status[1] = 2;		/* mark this down */
 			}
 			else
 			{
 				pool_log("starting degeneration. shutdown master host %s(%d)",
 						 pool_config.backend_host_name,
 						 pool_config.backend_port);
+				pool_config.server_status[0] = 2;		/* mark this down */
 			}
 		}
+
+		else if (pool_config.master_slave_enabled)
+		{
+			degenerated = 1;
+
+			if (sig == SIGUSR2)
+			{
+				pool_log("starting degeneration. shutdown slave host %s(%d)",
+						 pool_config.secondary_backend_host_name,
+						 pool_config.secondary_backend_port);
+				pool_config.server_status[1] = 2;		/* mark this down */
+			}
+			else
+			{
+				pool_log("starting degeneration. shutdown master host %s(%d)",
+						 pool_config.backend_host_name,
+						 pool_config.backend_port);
+				pool_config.server_status[0] = 2;		/* mark this down */
+			}
+		}
+
 		else if (!degenerated && pool_config.secondary_backend_port != 0)
 		{
 			pool_log("starting failover from %s(%d) to %s(%d)",
@@ -788,6 +835,7 @@ static RETSIGTYPE failover_handler(int sig)
 					   pool_config.current_backend_port,
 					   pool_config.secondary_backend_host_name,
 					   pool_config.secondary_backend_port);
+			pool_config.server_status[0] = 2;		/* mark this down */
 		}
 		else
 		{
@@ -822,6 +870,19 @@ static RETSIGTYPE failover_handler(int sig)
 				pool_config.current_backend_port = pool_config.secondary_backend_port;
 			}
 		}
+
+		else if (pool_config.master_slave_enabled)
+		{
+			/* disable master/slave mode */
+			pool_config.master_slave_enabled = 0;
+
+			if (sig == SIGUSR1)
+			{
+				pool_config.current_backend_host_name = pool_config.secondary_backend_host_name;
+				pool_config.current_backend_port = pool_config.secondary_backend_port;
+			}
+		}
+
 		else if (!degenerated && pool_config.secondary_backend_port != 0)
 		{
 			/* fail over to secondary */
@@ -844,7 +905,7 @@ static RETSIGTYPE failover_handler(int sig)
 		{
 			if (sig == SIGUSR2)
 			{
-				pool_log("degenration done. shutdown secondary host %s(%d)",
+				pool_log("degeneration done. shutdown secondary host %s(%d)",
 						 pool_config.secondary_backend_host_name,
 						 pool_config.secondary_backend_port);
 			}
