@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
-* $Header: /cvsroot/pgpool/pgpool/pool_stream.c,v 1.2 2005/06/13 13:29:37 t-ishii Exp $
+* $Header: /cvsroot/pgpool/pgpool/pool_stream.c,v 1.4 2005/07/30 12:30:20 t-ishii Exp $
 *
 * pgpool: a language independent connection pool server for PostgreSQL 
 * written by Tatsuo Ishii
@@ -33,6 +33,7 @@
 #include "pool.h"
 
 #define READBUFSZ 1024
+#define WRITEBUFSZ 8192
 
 static int mystrlen(char *str, int upper, int *flag);
 static int mystrlinelen(char *str, int upper, int *flag);
@@ -56,13 +57,15 @@ POOL_CONNECTION *pool_open(int fd)
 
 	memset(cp, 0, sizeof(*cp));
 
-	cp->write_fd = fdopen(fd, "w");
-	if (cp->write_fd == NULL)
+	/* initialize write buffer */
+	cp->wbuf = malloc(WRITEBUFSZ);
+	if (cp->wbuf == NULL)
 	{
-		pool_error("pool_open: fdopen failed: %s",strerror(errno));
-		free(cp);
+		pool_error("pool_open: malloc failed");
 		return NULL;
 	}
+	cp->wbufsz = WRITEBUFSZ;
+	cp->wbufpo = 0;
 
 	/* initialize pending data buffer */
 	cp->hp = malloc(READBUFSZ);
@@ -88,7 +91,8 @@ POOL_CONNECTION *pool_open(int fd)
 */
 void pool_close(POOL_CONNECTION *cp)
 {
-	fclose(cp->write_fd);
+	close(cp->fd);
+	free(cp->wbuf);
 	free(cp->hp);
 	if (cp->sbuf)
 		free(cp->sbuf);
@@ -136,6 +140,12 @@ int pool_read(POOL_CONNECTION *cp, void *buf, int len)
 		readlen = read(cp->fd, readbuf, READBUFSZ);
 		if (readlen == -1)
 		{
+			if (errno == EINTR || errno == EAGAIN)
+			{
+				pool_debug("pool_read: retrying due to %s", strerror(errno));
+				continue;
+			}
+
 			pool_error("pool_read: read failed (%s)", strerror(errno));
 
 			if (cp->isbackend)
@@ -151,13 +161,16 @@ int pool_read(POOL_CONNECTION *cp, void *buf, int len)
 		}
 		else if (readlen == 0)
 		{
-			pool_error("pool_read: EOF encountered");
-
 			if (cp->isbackend)
 			{
+				pool_error("pool_read2: EOF encountered with backend");
+				return -1;
+
+#ifdef NOT_USED
 			    /* fatal error, notice to parent and exit */
 			    notice_backend_error(!cp->issecondary_backend);
 				exit(1);
+#endif
 			}
 			else
 			{
@@ -244,6 +257,12 @@ char *pool_read2(POOL_CONNECTION *cp, int len)
 		readlen = read(cp->fd, buf, len);
 		if (readlen == -1)
 		{
+			if (errno == EINTR || errno == EAGAIN)
+			{
+				pool_debug("pool_read2: retrying due to %s", strerror(errno));
+				continue;
+			}
+
 			pool_error("pool_read2: read failed (%s)", strerror(errno));
 
 			if (cp->isbackend)
@@ -259,13 +278,16 @@ char *pool_read2(POOL_CONNECTION *cp, int len)
 		}
 		else if (readlen == 0)
 		{
-			pool_error("pool_read2: EOF encountered");
-
 			if (cp->isbackend)
 			{
+				pool_error("pool_read2: EOF encountered with backend");
+				return NULL;
+
+#ifdef NOT_USED
 			    /* fatal error, notice to parent and exit */
 			    notice_backend_error(!cp->issecondary_backend);
 				exit(1);
+#endif
 			}
 			else
 			{
@@ -284,34 +306,149 @@ char *pool_read2(POOL_CONNECTION *cp, int len)
 }
 
 /*
-* write len bytes from cp
+* write len bytes to cp the write buffer.
 * returns 0 on success otherwise -1.
 */
 int pool_write(POOL_CONNECTION *cp, void *buf, int len)
 {
-	if (!cp->no_forward)
-		fwrite(buf, len, 1, cp->write_fd);
+	int reqlen;
+
+	if (len < 0)
+	{
+		pool_error("pool_write: invalid request size: %d", len);
+		return -1;
+	}
+
+	if (cp->no_forward)
+		return 0;
+
+	/* check buffer size */
+	reqlen = cp->wbufpo + len;
+
+	if (reqlen > cp->wbufsz)
+	{
+		char *p;
+
+		reqlen = (reqlen/WRITEBUFSZ+1)*WRITEBUFSZ;
+		p = realloc(cp->wbuf, reqlen);
+		if (p == NULL)
+		{
+			pool_error("pool_write: realloc failed");
+			return -1;
+		}
+		cp->wbuf = p;
+		cp->wbufsz = reqlen;
+	}
+
+	memcpy(cp->wbuf+cp->wbufpo, buf, len);
+	cp->wbufpo += len;
 
 	return 0;
 }
 
 /*
-* flush write buffer
-*/
-int pool_flush(POOL_CONNECTION *cp)
+ * flush write buffer
+ */
+int pool_flush_it(POOL_CONNECTION *cp)
 {
+	int sts;
+	int wlen;
+	int offset;
+	wlen = cp->wbufpo;
+
+	if (wlen == 0)
+	{
+		return 0;
+	}
+
+	offset = 0;
+
 	for (;;)
 	{
 		errno = 0;
 
-		if (fflush(cp->write_fd) == 0)
-			break;
+#ifdef NOT_USED
+		if (!cp->isbackend)
+		{
+			fd_set	writemask;
+			fd_set	exceptmask;
 
-		if (errno == EAGAIN)
+			FD_ZERO(&writemask);
+			FD_ZERO(&exceptmask);
+			FD_SET(cp->fd, &writemask);
+			FD_SET(cp->fd, &exceptmask);
+
+			sts = select(cp->fd+1, NULL, &writemask, &exceptmask, NULL);
+			if (sts == -1)
+			{
+				if (errno == EAGAIN || errno == EINTR || errno == EWOULDBLOCK)
+					continue;
+
+				pool_error("pool_flush_it: select() failed. reason: %s", strerror(errno));
+				return -1;
+			}
+			else if (sts == 0)
+			{
+				continue;
+			}
+			else if (FD_ISSET(cp->fd, &exceptmask))
+			{
+				pool_log("pool_flush_it: exception occured");
+				return -1;
+			}
+		}
+#endif
+		sts = write(cp->fd, cp->wbuf + offset, wlen);
+
+		if (sts > 0)
+		{
+			wlen -= sts;
+
+			if (wlen == 0)
+			{
+				/* write completed */
+				break;
+			}
+
+			else if (wlen < 0)
+			{
+				pool_error("pool_flush_it: invalid write size %d", sts);
+				return -1;
+			}
+
+			else
+			{
+				/* need to write remaining data */
+				offset += sts;
+				continue;
+			}
+		}
+
+		else if (errno == EAGAIN || errno == EINTR)
+		{
 			continue;
+		}
 
-		pool_error("pool_flush: fflush failed (%s)", strerror(errno));
+		else
+		{
+			pool_error("pool_flush_it: write failed (%s) offset: %d wlen: %d",
+					   strerror(errno), offset, wlen);
+			return -1;
+		}
+	}
 
+	cp->wbufpo = 0;
+
+	return 0;
+}
+
+/*
+* flush write buffer and degenerate/failover if error occurs
+*/
+int pool_flush(POOL_CONNECTION *cp)
+{
+	if (pool_flush_it(cp) == -1)
+	{
 		if (cp->isbackend)
 		{
 			notice_backend_error(!cp->issecondary_backend);
@@ -448,6 +585,14 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 			    return NULL;
 			}
 		}
+		else if (readlen == 0)	/* EOF detected */
+		{
+			/*
+			 * just returns an error, not trigger failover or degeneration
+			 */
+			pool_error("pool_read_string: read () EOF detected");
+			return NULL;
+		}
 
 		/* check overrun */
 		if (line)
